# -*- python -*-
# $RCSfile: reporter.py,v $
# $Revision: 1.37 $
# $Author: reida $
# $Date: 2004/08/23 19:20:38 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Generic error-reporting machine.  Should be able to
# either simply log errors, or request a course of
# action from the user.

from SWIG.common import switchboard
from common import debug
from common import guitop
from common import thread_enable
from common.IO import questioner
import SWIG.common.lock
import string
import sys
import traceback


# List of allowed message classes.
messageclasses = ["Log", "Warning", "Report", "Query", "Error"]

# New message-handling class, which maintains a database of different
# types of message.  Will eventually have settings allowing for
# suppression of certain types of messages.
# In GUI mode, this thing gets a special window.

# It may be desirable to set an upper limit on the number of
# messages stored, so as not to chew up too much memory.

class MessageManager:
    lock = SWIG.common.lock.Lock()
    def __init__(self):
        # Set flags.  For each type of message, if the flag is set,
        # then the messages of that type should be displayed to the
        # user.
        self.flag_dict = {}
        for f in messageclasses:
            self.flag_dict[f]=1

        # Message list is a list of tuples, of the form (message, category).
        self.message_list = []
        self._pop_up_warnings = 1
        
        ## ProgressBar variables
        self.barup = 0
        self.bar_text=" "
        
    def get_warning_pop_up(self):
        return self._pop_up_warnings

    def set_warning_pop_up(self, arg=1):
        self._pop_up_warnings = arg
    
    def set_flag(self, flag, value):
        self.flag_dict[flag]=value

    def get_flag(self, flag):
        return self.flag_dict[flag]

    # "Smart write" routine -- the complexity originates in the need
    # to be able to write out messages (in text mode) while
    # partially-drawn text progress bars are in place.  The "barup"
    # flag indicates if there's a progress bar -- in that case, erase
    # it, write the message, then restore the progress bar and carry
    # on as before.  Called by both _append and display_bar, the
    # latter being the one that provides the bar_ID.
    def write(self, msg, bar_ID = None):
        if self.barup == 0 and bar_ID is None:
            ## if there is no bar up, then do the usual
            print msg
            sys.stdout.flush()
            return
        elif self.barup == 0 and bar_ID is not None:
            # TODO LATER: Should this raise an exception, or
            # is benign behavior in this circumstance important?
            return
        elif  self.barup == 1 and bar_ID is None:
            # This is the "smart" case -- erase the bar,
            # write the message, then redraw the bar by re-calling
            # self.display_bar.
            bar_len = len(self.bar_text)
            empty_text = bar_len*" " + "\r"
            sys.stdout.write(empty_text)
            sys.stdout.flush()

            print msg
            sys.stdout.flush()

            self.display_bar()
            return
        else:
            ## if I am drawing the actual bar, just draw it
            self.display_bar(msg)

    # All of the "normal" (i.e. non-progress-bar) messaging calls
    # eventually call "append", which adds the message to the
    # database.  This function also returns the message, although
    # most of the calls discard it, except "warning".
    def _append(self, type, *args):
        MessageManager.lock.acquire()
        message = string.join([str(x) for x in args], ' ')
        switchboard.notify( "write message", (message, type) )
        if debug.debug() and type!="Log":
            print message
        self.message_list.append( (message, type) )
        if self.flag_dict[type]:
            self.write(message)
        MessageManager.lock.release()
        return message
        
    def all_messages(self):
        return self.message_list
    
    def log(self, *args):
        self._append("Log", *args)
            
    def warn(self, *args):
        message = self._append("Warning", *args)
        switchboard.notify("messagemanager warning", message)

    def report(self, *args):
        self._append("Report", *args)

    # Arguments for this are different, this wraps the questioner.
    def query(self, question, *answers, **kwargs):
        ans = questioner.questioner(question, *answers, **kwargs)
        self._append("Query", question, ans)
        return ans

    def error(self, *args):
        sys.excepthook(*sys.exc_info())
        self._append("Error", *args)

    ##  progress bar controls
    def start_progressbar(self):
        if guitop.top() is None:
            self.barup = 1

    def end_progressbar(self):
        if guitop.top() is None:
            self.barup = 0
            bar_len = len(self.bar_text)
            empty_text = bar_len*" " + "\r"
            sys.stdout.write(empty_text)
            sys.stdout.flush()
        
    def display_bar(self, text = None):
        if text is not None:
            self.bar_text = text
        sys.stdout.write(self.bar_text)
        sys.stdout.flush() ## Do not remove this line. Without it,
        ## the text progress bar is never displayed.
        
# The one and only.
messagemanager = MessageManager()

# Convenience, make functions available at the module level.
def report(*args):
    messagemanager.report(*args)

def log(*args):
    messagemanager.log(*args)

# Warning: The warning function can bring up a pop-up, but does
# block in threads.  Blocking behavior can be provided by
# calling it with mainthread.runBlock.
def warn(*args):
    messagemanager.warn(*args)

# This function must block the child thread in thread mode,
# and so should be called with mainthread.runBlock by
# threaded menu items.
def query(question, *args, **kwargs):
    return messagemanager.query(question, *args, **kwargs)

def error(*args):
    messagemanager.error(*args)


# Over-ride-able function for generating a new display -- does nothing
# in command-line mode.
def _new_messages():
    pass

## progress bar module level functions
def start_progressbar():
    if guitop.top() is None:
        ## text progress bars should only be called
        ## in text mode
        messagemanager.start_progressbar()
    else:
        pass

def end_progressbar():
    if guitop.top() is None:
        ## text progress bars should only be called
        ## in text mode
        messagemanager.end_progressbar()
    else:
        pass

def progressbar(text = None):
    if guitop.top() is None:
        ## text progress bars should only be called
        ## in text mode
        messagemanager.write(text, 1)
    else:
        pass
    
def progressbar_completed():
    if guitop.top() is None:
        import time
        time.sleep(0.2) ## give a last chance to mainloop to show our completed work 
   
    
def progressbar_cleanup():
    pass ## sys.stdout.flush()

######################

# Object providing a file-like interface to the message manager's
# "report" function, useful for outputs which do not know/care what
# they're writing to.
class ReportFile:
    def __init__(self):
        self.buffer = ""
    def close(self):
        pass
    # Not quite file-like -- a call to "flush" will add a line separator
    # at the end of the flushed buffer data.
    def flush(self):
        global messagemanager
        if self.buffer:
            messagemanager.report(self.buffer)
        self.buffer=""
    def read(self):
        return ""
    def readline(self):
        return ""
    def write(self, data):
        global messagemanager
        self.buffer += data
        stringset = string.split(self.buffer, '\n')
        for s in stringset[:-1]:
            messagemanager.report(s)
        self.buffer = stringset[-1]

fileobj = ReportFile()

