# -*- python -*-
# $RCSfile: cregisteredclass.py,v $
# $Revision: 1.13 $
# $Author: langer $
# $Date: 2004/09/07 15:17:00 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


# The registerCClass() function makes a swigged C++ class act like
# RegisteredClass.  For documentation on the individual functions, see
# registeredclass.py, where the Python RegisteredClass class is
# defined.

# The Registrations for the subclasses of a C++ registered class must
# be instances of the Registration class in this file
# (cregisteredclass.py) rather than the one in registeredclass.py.

from SWIG.common import ooferror
from SWIG.common import timestamp
from common import debug
from common import registeredclass
from common import utils
import string

def registerCClass(klass):
    klass.registry = []

    if not hasattr(klass, 'getRegistrationIndex'):
        def getRegistrationIndex(self):
            for i in range(len(self.registry)):
                if isinstance(self, self.registry[i].subclass) or \
                   issubclass(self.registry[i].subclass, self.__class__):
                    return i
        klass.getRegistrationIndex = getRegistrationIndex

    if not hasattr(klass, 'getNonSecretRegistrationIndex'):
        def getNonSecretRegistrationIndex(self):
            count = 0
            for registration in self.registry:
                if isinstance(self, registration.subclass):
                    return count
                if not registration.secret:
                    count += 1
        klass.getNonSecretRegistrationIndex = getNonSecretRegistrationIndex

    if not hasattr(klass, 'getRegistration'):
        def getRegistration(self):
            return self.registry[self.getRegistrationIndex()]
        klass.getRegistration = getRegistration

    if not hasattr(klass, 'getParamValues'):
        def getParamValues(ego):
            return ego._keepargs
        klass.getParamValues = getParamValues

    if not hasattr(klass, 'setDefaultParams'):
        def setDefaultParams(self):
            registration = self.getRegistration()
            registration.setDefaultParams(self.getParamValues())
        klass.setDefaultParams = setDefaultParams

    if not hasattr(klass, 'getDefaultParams'):
        def getDefaultParams(self):
            return self.getRegistration().params
        klass.getDefaultParams = getDefaultParams

    if not hasattr(klass, 'paramrepr'):
        def paramrepr(self):
            values = self.getParamValues()
            names = [p.name for p in self.getRegistration().params]
            return string.join(['%s=%s' % (name, value)
                                for (name, value) in zip(names, values)], ',')
        klass.paramrepr = paramrepr

    if not hasattr(klass, '__eq__'):
        def __eq__(self, other):
            if other is None:
                return 0
            if self.__class__ != other.__class__:
                return 0
            reg = self.getRegistration()
            if reg.to_base:
                self_base = self.to_base
                other_base = other.to_base()
                for k in self_base.__dict__.keys():
                    if getattr(self_base, k) != getattr(other_base, k):
                        return 0
            else:
                for i in self.__dict__.keys():
                    if getattr(self, k) != getattr(other, k):
                        return 0
            return 1
        klass.__eq__ = __eq__

    if not hasattr(klass, '__ne__'):
        def __ne__(self, other):
            return 1 - self.__eq__(other)
        klass.__ne__ = __ne__
    

class Registration(registeredclass.Registration):
    def __init__(self, name, regclass, subclass, ordering,
                 params=[], secret=0, **kwargs):

        registeredclass.Registration.__init__(self, name, regclass, subclass,
                                              ordering, params, secret,
                                              **kwargs)
        # __repr__ has to be redefined here, instead of in the
        # baseclass, so that it overrides the swig-generated __repr__
        # in the subclass.
        def repr(ego):
            return '%s(%s)' % (ego.__class__.__name__, ego.paramrepr())
        self.subclass.__repr__ = repr
        self.subclass.__str__ = repr

        # Actually, __repr__ has to be redefined in the xxxxPtr
        # version of the subclass, which is the parent of the class we
        # have.  The repr for this class, however, should *not* print
        # the 'Ptr' part of the class name, since that name should be
        # hidden from the user, and anyhow there's no way to construct
        # xxxxPtr objects directly from a script.
        
        def repr(ego):
            return '%s(%s)' % (ego.__class__.__name__[:-3], ego.paramrepr())

        # Just in case there is more than one base class, check for
        # the one that's derived from regclass.
        for base in self.subclass.__bases__:
            if issubclass(base, self.registeredclasses[0]):
                base.__repr__ = repr
                return

    def substituteClass(self, newclass):
        # Sometimes a module is loaded that redefines the behavior of
        # an existing RegisteredClass subclass, for example by adding
        # graphics routines to it.  If the subclass is a Python class,
        # that's no problem, but changing a C++ class at run time is
        # more difficult.  The substituteClass() routine allows a
        # derived class to be substituted for a previously registered
        # C++ class.  The underlying C++ version of the new class must
        # be a subclass of the C++ version of the old class, although
        # this isn't checked for (since it would be a pain to do.. we
        # only have the non-Ptr swigged classes here).  Note that the
        # repr of the class is *not* changed, so scripted commands
        # still refer to the old class.  This is the correct behavior
        # when the modification class is just adding graphics
        # capability.
        oldclass = self.subclass
        self.subclass = newclass
        def repr(ego, oldclass=oldclass):
            return '%s(%s)' % (oldclass.__name__, ego.paramrepr())
        self.subclass.__repr__ = repr
        self.subclass.__str__ = repr
        
    def exportToOOF(self):
        # C Registered classes have to export their Registrations,
        # rather than their subclasses, so that all constructor calls
        # go through Registration.__call__.  This gives the
        # Registration control over the scope of the arguments.  See
        # __call__, below.
        utils.OOFdefine(self.subclass.__name__, self)
        
    def __call__(self, *args, **kwargs):
        for arg, param in zip(args, self.params):
            param.value = arg
        for p in self.params:
            try:
                p.value = kwargs[p.name]
            except KeyError:
                pass
        argvals = [p.value for p in self.params]
        object = self.subclass(*argvals)
        
        # Keep an extra reference to the arguments to ensure that the
        # Python garbage collector won't destroy them before the
        # registered class object itself is destroyed.  Also, this
        # allow getParamValues to return the original Python form of
        # the arguments.  If getParamValues goes into C++ to return
        # the arguments, it will return the Ptr form, which will
        # confuse the RegisteredClass type checking.
        object._keepargs = argvals

        object.timestamp = timestamp.TimeStamp()
        return object
