// -*- C++ -*-
// $RCSfile: field.C,v $
// $Revision: 1.32 $
// $Author: langer $
// $Date: 2004/07/21 19:14:09 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include "common/switchboard.h"
#include "engine/elementnodeiterator.h"
#include "engine/field.h"
#include "engine/indextypes.h"
#include "engine/femesh.h"
#include "engine/node.h"
#include "engine/ooferror.h"
#include "engine/outputval.h"
#include "engine/property.h"
#include "common/removeitem.h"
#include "common/tostring.h"
#include "common/trace.h"
#include <iostream>
#include <vector>
using namespace std;

vector<Field*> &Field::allfields() {
  static vector<Field*> all_fields;
  return all_fields;
}

vector<CompoundField*> &CompoundField::allcompoundfields() {
  static vector<CompoundField*> all_fields;
  return all_fields;
}

CompoundField *getCompoundFieldByIndex(int index) {
  return CompoundField::allcompoundfields()[index];

}

int countFields() {
  return Field::allfields().size(); 
}

int countCompoundFields() {
  return CompoundField::allcompoundfields().size();
}

Field *Field::getField(const string &nm) {
  const vector<Field*> &list = allfields();
  for(int i=0; i<list.size(); i++) {
    if(list[i]->name() == nm) {
      return list[i];
    }
  }
  throw ErrProgrammingError("Unknown Field \"" + nm + "\"",
			    __FILE__, __LINE__);
}

Field::Field(const string &nm, int dofs)
  : name_(nm),
    dim(dofs),
    index_(allfields().size()),
    in_plane_(true)
{
  allfields().push_back(this);
}


Field::~Field() {
  // this should be called only if the program is exiting... Fields
  // are global static objects and aren't deleted
  remove_item(allfields(), this);
}

ostream &operator<<(ostream &os, const Field &field) {
  os << field.name();
  return os;
}

const std::string &Field::name() const {
  return name_;
}

//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//

void Field::define(FEMesh *mesh) {
  mesh->do_define_field(*this);
}

void Field::undefine(FEMesh *mesh) {
  mesh->do_undefine_field(*this);
}

bool Field::is_defined(const FEMesh *mesh) const {
  return mesh->is_defined_field(*this);
}

// initialization of a Field with a C++ function
void Field::initialize(FEMesh *mesh, double (*f)(const Coord &position, int dof))
{
  mesh->define_field(*this);
  for(FuncNodeIterator i(mesh); !i.end(); ++i)
  {
    for(int d = 0; d<dim; ++d)
    {
      operator()(mesh, (*mesh)[i],d)->value() = (*f)((*mesh)[i]->position(),d);
    }
  }
}

// initialization of a Field with a Python function
void Field::initialize(FEMesh *mesh, double (*f)(const Coord &, int, void *),
		       void *clientdata)
{
  mesh->define_field(*this);
  for(FuncNodeIterator i(mesh); !i.end(); ++i)
  {
    for(int d = 0; d<dim; ++d)
    {
      operator()(mesh, (*mesh)[i],d)->value() =
	(*f)((*mesh)[i]->position(),d,clientdata);
    }
  }
}

//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//

void Field::activate(FEMesh *mesh) {
  mesh->do_activate_field(*this);
}

void Field::deactivate(FEMesh *mesh) {
  mesh->do_deactivate_field(*this);
}

bool Field::is_active(const FEMesh *mesh) const {
  return mesh->is_active_field(*this);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void Field::registerProperty(Property *prop) const {
  prop->require_field(*this);
}

void CompoundField::registerProperty(Property *prop) const {
  prop->require_field(*this);
  prop->require_field(*zfield_);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// position of a given component in the dof lists in the nodes

int Field::localindex(const FEMesh *mesh, const FieldIndex &component) const {
  return mesh->defined_index(*this) + component.integer();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

CompoundField::CompoundField(const string &name, int dim, Field *follies)
  : Field(name, dim),
    zfield_(follies),
    cfield_indx(allcompoundfields().size())
{
  zfield_->set_oop(); // Set z component to be out-of-plane.
  allcompoundfields().push_back(this);
}

CompoundField::~CompoundField() {
  remove_item(allcompoundfields(), this);
}

void CompoundField::set_in_plane(FEMesh *mesh, bool planarity) {
  mesh->set_in_plane(*this, planarity);
  // activate the out of plane field only if this field is active and
  // not in plane.
  if(in_plane(mesh))
    mesh->do_deactivate_field(*zfield_);
  else {
    if(mesh->is_active_field(*this)) {
      mesh->do_activate_field(*zfield_);
    }
  }
}

bool CompoundField::in_plane(const FEMesh *mesh) const {
  return mesh->in_plane(*this);
}

void CompoundField::define(FEMesh *mesh) {
  mesh->do_define_field(*this);
  mesh->do_define_field(*zfield_);
}

void CompoundField::undefine(FEMesh *mesh) {
  mesh->do_undefine_field(*this);
  mesh->do_undefine_field(*zfield_);
}

void CompoundField::activate(FEMesh *mesh) {
  mesh->do_activate_field(*this);
  if(!in_plane(mesh))
    mesh->do_activate_field(*zfield_);
}

void CompoundField::deactivate(FEMesh *mesh) {
  mesh->do_deactivate_field(*this);
  mesh->do_deactivate_field(*zfield_);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

ScalarField::ScalarField(const string &nm)
  : Field(nm, 1),
    ScalarFieldBase(nm),
    CompoundField(nm, 1, new ScalarFieldBase(nm+string("_z")))
{
}

const string ScalarField::classname_("ScalarField");

DegreeOfFreedom *ScalarFieldBase::operator()(const FEMesh *mesh,
					     const FuncNode *node) const
{
  return node->doflist[mesh->defined_index(*this)];
}

DegreeOfFreedom *ScalarFieldBase::operator()(const FEMesh *mesh,
					     const ElementFuncNodeIterator &ei)
  const 
{
  return operator()(mesh, ei.funcnode());
}

// These two silly versions have to exist because these two functions
// have to be virtual in the Field base class. That's because
// CompoundField::out_of_plane returns a pointer to a Field base class
// object.
DegreeOfFreedom *ScalarFieldBase::operator()(const FEMesh *mesh,
					     const FuncNode *node,
					     int component) const
{
  assert(component == 0);
  return node->doflist[mesh->defined_index(*this)];
}

DegreeOfFreedom *ScalarFieldBase::operator()(const FEMesh *mesh,
					     const ElementFuncNodeIterator &ei,
					     int component) const 
{
  assert(component == 0);
  return operator()(mesh, ei.funcnode());
}

OutputValue ScalarFieldBase::newOutputValue() const {
  return OutputValue(new ScalarOutputVal(0.));
}

OutputValue ScalarFieldBase::output(const FEMesh *mesh, const FuncNode &node)
  const
{
  return OutputValue(new ScalarOutputVal(operator()(mesh, node)->value()));
}

OutputValue ScalarFieldBase::output(const FEMesh *mesh,
				    const ElementFuncNodeIterator &node)
  const
{
  return OutputValue(new ScalarOutputVal(operator()(mesh, node)->value()));
}

void ScalarFieldBase::setValueFromOutputValue(const FEMesh *mesh,
					      const FuncNode &node,
					      const OutputValue *ov)
{
  const ScalarOutputVal &sov =
    dynamic_cast<const ScalarOutputVal&>(ov->value());
  operator()(mesh, node)->value() = sov.value();
}


IteratorP ScalarFieldBase::iterator(Planarity) const {
  return IteratorP(new ScalarFieldIterator);
}

IndexP ScalarFieldBase::componenttype() const {
  return IndexP(new ScalarFieldIndex);
}

IndexP ScalarFieldBase::getIndex(const std::string &) const {
  return IndexP(new ScalarFieldIndex);
}

//------------------------------

TwoVectorField::TwoVectorField(const string &nm)
  : Field(nm, 2),
    TwoVectorFieldBase(nm),
    CompoundField(nm, 2, new VectorFieldBase(nm+string("_z"), 3))
{
}

const string TwoVectorField::classname_("TwoVectorField");

DegreeOfFreedom *TwoVectorFieldBase::operator()(const FEMesh *mesh,
						const FuncNode *node,
						int comp) const
{
  assert(comp >= 0 && comp < 2);
  return node->doflist[mesh->defined_index(*this) + comp];
}

DegreeOfFreedom *
TwoVectorFieldBase::operator()(const FEMesh *mesh, 
			       const ElementFuncNodeIterator &ei,
			       int comp) const
{
  return operator()(mesh, ei.funcnode(), comp);
}

OutputValue TwoVectorFieldBase::newOutputValue() const {
  return OutputValue(new VectorOutputVal(2));
}

OutputValue TwoVectorFieldBase::output(const FEMesh *mesh,
				       const FuncNode &node) const
{
  VectorOutputVal *vov = new VectorOutputVal(2);
  OutputValue ov(vov);
  (*vov)[0] = operator()(mesh, node, 0)->value();
  (*vov)[1] = operator()(mesh, node, 1)->value();
  return ov;
}

OutputValue TwoVectorFieldBase::output(const FEMesh *mesh,
				       const ElementFuncNodeIterator &node)
  const 
{
  VectorOutputVal *vov = new VectorOutputVal(2);
  OutputValue ov(vov);
  (*vov)[0] = operator()(mesh, node, 0)->value();
  (*vov)[1] = operator()(mesh, node, 1)->value();
  return ov;
}

void TwoVectorFieldBase::setValueFromOutputValue(const FEMesh *mesh, 
						 const FuncNode &node,
						 const OutputValue *ov)
{
  const VectorOutputVal &vov = 
    dynamic_cast<const VectorOutputVal&>(ov->value());
  (*this)(mesh, node, 0)->value() = vov[0];
  (*this)(mesh, node, 1)->value() = vov[1];
}

// The TwoVectorHelper class allows TwoVectorField(FuncNode) to act like a
// Coord.
TwoVectorHelper TwoVectorFieldBase::operator()(const FEMesh *mesh,
					       const FuncNode *node) const
{
  int idx = mesh->defined_index(*this);
  return TwoVectorHelper(node->doflist[idx], node->doflist[idx+1]);
}

TwoVectorHelper
TwoVectorFieldBase::operator()(const FEMesh *mesh,
			       const ElementFuncNodeIterator &ei) const
{
  return operator()(mesh, ei.funcnode());
}

// field(node) = coord
TwoVectorHelper &TwoVectorHelper::operator=(const Coord &v) {
  x->value() = v(0);
  y->value() = v(1);
  return *this;
}

// coord = field(node)
TwoVectorHelper::operator const Coord() const {
  return Coord(x->value(), y->value());
}

IteratorP TwoVectorFieldBase::iterator(Planarity) const {
  return IteratorP(new VectorFieldIterator(0, 2));
}

IndexP TwoVectorFieldBase::componenttype() const {
  return IndexP(new VectorFieldIndex);
}

IndexP TwoVectorFieldBase::getIndex(const std::string &str) const {
  return IndexP(new VectorFieldIndex(str[0] - 'x'));
}

//-------------------------------

DegreeOfFreedom *VectorFieldBase::operator()(const FEMesh *mesh,
					     const FuncNode *node,
					     int comp) const
{
  assert(comp >= 0 && comp < dim);
  return node->doflist[mesh->defined_index(*this) + comp];
}

DegreeOfFreedom *VectorFieldBase::operator()(const FEMesh *mesh,
					     const ElementFuncNodeIterator &ei,
					     int comp)
  const
{
  return operator()(mesh, ei.funcnode(), comp);
}

OutputValue VectorFieldBase::newOutputValue() const {
  return OutputValue(new VectorOutputVal(dim));
}

OutputValue VectorFieldBase::output(const FEMesh *mesh, const FuncNode &node)
  const 
{
  VectorOutputVal *vov = new VectorOutputVal(dim);
  OutputValue ov(vov);
  for(int i=0; i<dim; i++) {
    (*vov)[i] = operator()(mesh, node, i)->value();
  }
  return ov;
}

OutputValue VectorFieldBase::output(const FEMesh *mesh,
				    const ElementFuncNodeIterator &node)
  const 
{
  VectorOutputVal *vov = new VectorOutputVal(dim);
  OutputValue ov(vov);
  for(int i=0; i<dim; i++) {
    (*vov)[i] = operator()(mesh, node, i)->value();
  }
  return ov;
}

void VectorFieldBase::setValueFromOutputValue(const FEMesh *mesh,
					      const FuncNode &node,
					      const OutputValue *ov)
{
  const VectorOutputVal &vov = 
    dynamic_cast<const VectorOutputVal&>(ov->value());
  for(int i=0; i<dim; i++)
    (*this)(mesh, node, i)->value() = vov[i];
}

// The VectorHelper class allows VectorField(FuncNode) to act like a Vector.
VectorHelper VectorFieldBase::operator()(const FEMesh *mesh, const FuncNode *node)
  const
{
  return VectorHelper(&(node->doflist[mesh->defined_index(*this)]), dim);
}

VectorHelper VectorFieldBase::operator()(const FEMesh *mesh,
					 const ElementFuncNodeIterator &ei)
  const
{
  return operator()(mesh, ei.funcnode());
}

// field(node) = vector
VectorHelper &VectorHelper::operator=(const VECTOR_D &v) {
  assert(v.dim() == ndof);
  for(int i=0; i<ndof; i++)
    x[i]->value() = v[i];
  return *this;
}

// vector = field(node)
VectorHelper::operator const VECTOR_D() const {
  VECTOR_D v(ndof);
  for(int i=0; i<ndof; i++)
    v[i] = x[i]->value();
  return v;
}

IteratorP VectorFieldBase::iterator(Planarity planarity) const {
  int mindim = (planarity == OUT_OF_PLANE ? 2 : 0);
  int maxdim = (planarity == IN_PLANE ? 2 : ndof());
  return IteratorP(new VectorFieldIterator(mindim, maxdim));
}

IndexP VectorFieldBase::componenttype() const {
  return IndexP(new VectorFieldIndex);
}

IndexP VectorFieldBase::getIndex(const std::string &str) const {
  return IndexP(new VectorFieldIndex(str[0] - 'x'));
}

