# -*- python -*-
# $RCSfile: optimize.py,v $
# $Revision: 1.3 $
# $Author: redwing $
# $Date: 2004/09/23 14:56:03 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from common import debug
from common import primitives
from common import registeredclass
from common.IO import reporter
from common.IO import parameter
from engine import fiddlenodesbase
from engine import skeletonmodifier
from SWIG.common.IO import progressbar
from SWIG.common import devel
from SWIG.common import switchboard
from SWIG.engine import preconditioner
from SWIG.engine import solver
from SWIG.engine import solverstepper
import random

FiddleNodes = fiddlenodesbase.FiddleNodes
FiddleNodesTargets = fiddlenodesbase.FiddleNodesTargets
IterationManager = fiddlenodesbase.IterationManager

class OptimizeMovePosition(fiddlenodesbase.FiddleNodesMovePosition):
    def __init__(self, delta):
        self.delta = delta
    def __call__(self, skeleton, node):
        px, py = skeleton.MS.sizeOfPixels()
        dx = random.gauss(0.0, self.delta*px)
        dy = random.gauss(0.0, self.delta*py)
        return primitives.Point(node.position()[0]+dx,
                                node.position()[1]+dy)

registeredclass.Registration(
    'Optimize Move Position', fiddlenodesbase.FiddleNodesMovePosition,
    OptimizeMovePosition,
    ordering=0,
    params = [parameter.FloatParameter("delta", 1.0, tip='width of gaussian distribution of node motions, in pixel units')],)

class Optimize(FiddleNodes, skeletonmodifier.SkeletonModifier):
    def __init__(self, criterion, iteration):
        T = 0
        targets = None
        FiddleNodes.__init__(self, targets, criterion, T, iteration)
        self.intro = "preparing to optimize skeleton...      "
        self.header = "optimizing skeleton: "
        self.outro = "optimization done "
    def postProcess(self, context):
        ## This function should first, create a mesh with the custom-made properties,
        ## second, assign "temporary" properties to pixels
        ## specify BCs, equations.
        ## third, iterate using the NR solver
        ## if solution is found, accept node moves and redraw skeleton
        ## repeat above steps until iteration criteria has been met
        ## or until NR solver fails
        ## finally, trash the mesh, and clean-up all the temporary objects
        skeleton = context.getObject()
        pBar = progressbar.getcProgressBar()
        self.count = 0
        pBar.set_message(self.header)
        before = skeleton.energyTotal(self.criterion.alpha)
        n = self.iteration.iterations
        ## create temporary FE mesh
        mesh = self.create_mesh(context)
        while self.iteration.goodToGo():
            self.count += 1
            # the context to acquires the writing persmissions
            # inside coreProcess.

            ## context thread-safe area starts here
            context.begin_writing()
            try:
                ## coreProcess sets its own set of locks 
                self.coreProcess(mesh) ## coreProcess solves a NR problem for an existing mesh
                skeleton.timestamp.increment()
            finally:
                context.end_writing()
            ## context thread-safe area ends here
                switchboard.notify("redraw") ## should redraw every sucessful iteration

                
            self.updateIteration()
            if pBar.query_stop():
                pBar.set_failure()
                pBar.set_message("Failed")
                break
            else:
                if self.pbar_type == "continuous":
                    pBar.set_progress(1.0*self.count/n)
                    pBar.set_message("%s%d/%d"
                                     % (self.header, self.count, n))

        switchboard.notify("skeleton nodes moved", context)
        if pBar.query_stop(): # or pBar.get_success() <0:
            pBar.set_failure()
            pBar.set_message("Failed")
            return 
        after = skeleton.energyTotal(self.criterion.alpha)
        if before:
            rate = 100.0*(before-after)/before
        else:
            rate = 0.0
        diffE = after - before
        reporter.report("%s deltaE = %10.4e (%6.3f%%)"
                        % (self.outro, diffE, rate))

                
    def coreProcess(self, mesh, activenodes = None):
        ## mesh is a REAL mesh
        pBar = progressbar.getcProgressBar()
        linear_solver = solver.GMRESSolver(1000, 200, 1e-8, preconditioner.BlockPreconditioner(ILU()))
        newton_solver = solverstepper.NewtonSolver(linear_solver, 1e-8, 20)
        newton_solver.apply_realmesh(mesh)
        ## should transfer the new node positions to the skeleton at this point
        
       
            

#################################################
if devel.devel():
    registeredclass.Registration('Optimize',
                                 skeletonmodifier.SkeletonModifier,
                                 Optimize,
                                 ordering=200,
                                 params=[
        parameter.Parameter('criterion', skeletonmodifier.SkelModCriterion,
                            tip = 'Acceptance criterion'),
        parameter.RegisteredParameter('iteration', IterationManager,
                                      tip='Iteration method')
        ],
                                 tip='Move large number of nodes by optimizing a global energy functional (by oofing our way through)'
                                 )
    
