# -*- python -*-
# $RCSfile: propertyregistration.py,v $
# $Revision: 1.76 $
# $Author: reida $
# $Date: 2004/09/16 18:55:45 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

# This file defines the PropertyRegistration class, which is used
# to get information on how to construct instances of the registered
# property classes when no such instances exist.

# Also, this file defines/contains the global AllProperties object,
# which looks like a dictionary, but has the capability of notifying
# the world when something has been inserted -- this is so the
# GUI PropertyTree can be updated.


from SWIG.common import ooferror
from SWIG.common import switchboard
from SWIG.engine import properties
from common import debug
from common import utils
from common.IO import mainmenu
from common.IO import oofmenu
from common.IO import parameter
from common.IO import reporter
from common.labeltree import LabelTree
import string, types


# Load the convertible registered types used as parameters
# by the properties.
from engine.IO import isocijkl
from engine.IO import orientationmatrix

# The PropertyManager is the object over which the GUI's propertyTree
# is built.  It has as data a labeltree whose objects are property
# registration entries.  There are two kinds of leaf entities on the
# tree, those corresponding to imported property classes and hosting a
# "generic" (or "unnamed") potential-instance, and their immediate
# children, are named instances of the property class corresponding to
# their parent.

OOF = mainmenu.OOF

OOF.addItem(oofmenu.OOFMenuItem('Property',cli_only=1))
OOF.LoadData.addItem(oofmenu.OOFMenuItem('Property'))

# MenuKey objects are used to identify the various menus that are
# automatically created from the Property LabelTree.
class MenuKey:
    def __init__(self, string, object):
        self.string = string
        self.object = object

class PropertyManager:
    def __init__(self):
        self.materialmanager = None # Set by materialmanager when it starts up.
        self.data = LabelTree()

        # Need to own the references to the key objects, which are
        # only weakly referenced in LabelTree, and used to identify
        # menus.  This allows the menus to disappear automatically
        # when the PropertyManager is destroyed.  (Why bother?  The
        # PropertyManager isn't destroyed until the program quits.)
        self.parametrizekey = MenuKey('Parametrize', self)
        self.loadkey = MenuKey('Load', self)


        def paramcopy(object):
            if object:
                # If this routine *doesn't* copy the parameters, but
                # instead uses them directly, then the menuitems
                # created by LabelTree.makeOOFMenu will use the
                # parameters in the PropertyRegistrations, and there's
                # no need to synchronize the default values of the
                # parameters in the registrations and the menus.
                # PropertyRegistrations are used to create two menus,
                # OOF.LoadData.Propery and OOF.Property.Parametrize, so
                # synchronization is an issue.
                return object.getDefaultParams()
##                return [ p.clone() for p in object.getDefaultParams() ]
            return []

        # Menu items are put into OOF.Property for scripts and
        # OOF.LoadData.Property for data files.  The menus are created
        # automatically by the LabelTree, which uses the given key to
        # distinguish them.
        OOF.Property.addItem(
            self.data.makeOOFMenu(name='Parametrize',
                                  key=self.parametrizekey,
                                  param_func = paramcopy,
                                  callback = self.parametrizercallback))

        OOF.LoadData.addItem(
            self.data.makeOOFMenu(name='Property',
                                  key=self.loadkey,
                                  param_func = paramcopy,
                                  callback = self.creatorcallback,
                                  params = [
            parameter.StringParameter('name')] ))

    # __setitem__ gets called whenever a new registration is
    # instantiated.  This happens for unnamed properties at start-up
    # time and for named properties whenever they're created.  Does
    # collision detection by first attempting a retrieval on the key
    # -- if that succeeds, you drop through to the "else:".
    def __setitem__(self,key,value):
        try:
            collision = self.data[key]
        except KeyError:
            # Key not found, OK to insert. 
            propclass = string.split(key,':')[-1]

            # Add it to the labeltree.
            self.data.__setitem__(key, value, ordering=value.ordering)
        else:
            # Name-collision has occurred.  This is only an error
            # if the parameter values conflict.
            for (p1, p2) in zip(collision.object.params, value.params):
                if (p1.name != p2.name) or (p1.value != p2.value):
                    raise KeyError(
                        "Assignment collision in PropertyManager, key %s."
                        % `key`)


    # Make a named instance of the given property.  Called by
    # OOF.Property.Copy, so it's not necessary to do all the checks
    # done in __setitem__.  The passed-in newname is the leaf
    # component, but oldname is a fully qualified name.
    def new_prop(self, oldname, newname):
        oldreg = self.data[oldname].object
        if hasattr(oldreg , "parent"):
            namelist = string.split(oldname, ':')[:-1]
            fullname = string.join(namelist+[newname], ':')
        else:
            fullname = oldname+":"+newname
        newreg = oldreg.named_copy(fullname)
        switchboard.notify("new property", newreg)
    
    def __getitem__(self,key):
        return self.data[key].object

    # Delete the named item.  LabelTree ("self.data") will raise
    # ErrUserError if the named item is a non-leaf.
    def delete(self, name):
        reg = self.data[name].object
        self.materialmanager.delete_prop(name)
        reg.unregister()
        self.data.delete(name)

    # Property name uniquifier, used when copying.  "oldprop" is a
    # fully qualified name of an existing property, and newname is an
    # unqualfied candidate "leaf" name.
    def uniqueName(self, oldprop, newname):
        oldreg = self.data[oldprop].object
        try:
            contextreg = oldreg.parent
        except AttributeError:
            contextreg = oldreg
        # Get the labeltreenode.
        ltn = self.data[self.data.objpath(contextreg)]
        nameset = ltn.children()
        return utils.menUniqueName(newname, nameset)
        

    # This menu callback gets called when the parameters of a
    # property are changed.  Figure out which property registration
    # object is relevant, and run its new_params() method.
    def parametrizercallback(self, menuitem, **kwargs):
        # Translate the menuitem's path to the tree's path.  The first
        # three words of the path are OOF.Property.Parametrize.
        # We want the remainder.
        treepath = string.split(menuitem.path(),".")[3:]
        reg = self.data[treepath].object
##        reg.setDefaultParamsFromParams(menuitem.params)
        reg.new_params()
        switchboard.notify("redraw")
        

    # Menu callback for OOF.LoadData.Property.  "reg" is the
    # registration entry corresponding to the menu item that initiated
    # the callback, e.g.
    # "OOF.LoadData.Property.Create.Elasticity.IsoElasiticty".  The
    # PropertyRegistration and NamedPropertyRegistration's writeData
    # methods ensure that "reg" is always a PropertyRegistration.
    def creatorcallback(self, menuitem, **kwargs):
        treepath = string.split(menuitem.path(),".")[3:]
        reg = self.data[treepath].object
        name = menuitem.params[0].value

        # Collision looks for the name under this tree, and if it
        # finds it, checks if the parameter values are all equal.  If
        # a collision occurs and the parameters conflict, an exception
        # is raised.  If collision returns "true", that means a
        # collision occurred but the parameters matched.  If collision
        # returns "false", a collision did not occur.
        namecollision, parametercollision = reg.collision(name,
                                                          menuitem.params[1:])
        if namecollision:               # name is a duplicate
            if parametercollision:      # parameters disagree
                if name != "":
                    raise ooferror.ErrSetupError("Named property in datafile conflicts with existing property '%s'" % name)
                # reparametrization of unnamed property
                if reg.materials:
                    raise ooferror.ErrSetupError("Unnamed property in datafile conflicts with existing property '%s'" % name)
                # Unnamed property is being reparametrized, but it's
                # not used in any materials, so it's ok.
                reg.new_params()
                switchboard.notify("redraw")
            # A name collision with no parameter collision doesn't
            # require any action.  The data file contained a property
            # identical to an existing property.
        else:
            # No collision, we must create a new NamedRegistration.
            # We know it's a NamedRegistration because unnamed
            # property registrations *always* produce a name conflict.
            fullname = string.join( treepath + [name], ":")
            newreg = reg.named_copy(fullname, menuitem.params[1:])
            switchboard.notify("redraw")
            

    

# AllProperties is the global instance of the manager.
AllProperties = PropertyManager()
utils.OOFdefine("AllProperties", AllProperties)

# Property registrations live in the tree data structure of the
# PropertyManager, and consequently are managed separately from the
# usual OOF Registration mechanism.  They do their own tree-based
# collision detection, and are not known to the global OOF namespace.
# Properties are instanced via the "OOF.Property.Create..." menu in
# scripts, and parametrized via the "OOF.Property.Parametrize..." 
# menu.  PropertyRegistration and NamedPropertyRegistration objects
# differ in that the former construct a default name for themselves,
# whereas the latter accept one as an argument.  Both prepend the
# resulting name as a parameter to their parameter list.

class PropertyRegistrationParent:
    def __init__(self, subclass, modulename, ordering):
        self.subclass = subclass        # for PythonExportability of Property
        self.modulename = modulename    # ditto
        self.ordering = ordering
        properties.Property.registry.append(self)

    def unregister(self):
        del properties.Property.registry[self.getIndex()]

    def getIndex(self):
        for i in range(len(properties.Property.registry)):
            if properties.Property.registry[i] is self:
                return i

    
# PropertyRegistration knows special things about itself, like the
# fact that all properties have names as their first arguments.
class PropertyRegistration(PropertyRegistrationParent):
    def __init__(self, name, subclass, modulename, ordering, params=[],
                 fields=[], fluxes=[], propertyType=None,
                 outputs=[]):

        PropertyRegistrationParent.__init__(self, subclass,
                                            modulename, ordering)

        # Save the fully-qualified name for error reporting.  This
        # datum should not be confused with the parameter "name",
        # which contains only the leaf part of the FQN.
        self._name = name
        # Keep a copy of the local parameters.
        self.params = params
        self._fields = fields           # fields required by this Property
        self._fluxes = fluxes           # fluxes to which it contributes
        self._outputs = outputs         # names of PropOutputs it contributes to

        if propertyType is None:
            raise ooferror.ErrPyProgrammingError(
                "Missing propertyType in PropertyRegistration %s" % name)
        self._propertyType=propertyType

        # name-indexed dictionary of materials in which this property
        # occurs -- it is managed by the MaterialProps objects, except
        # in the new_params call, where it is local.
        self.materials = {}
        # At creation time, all parameters are new.
        self.new_params()

        # Add yourself to the AllProperties data object.  "name"
        # must be the fully qualified name of the property.
        AllProperties[name]=self

    # This returns the fully-qualified name in the labeltree.
    def name(self):
        return self._name

    def baseName(self):                 # different in NamedPropertyRegistration
        return self._name
    
    # These functions are different in the NamedPropertyRegistration
    def fields(self):
        return self._fields
    def fluxes(self):
        return self._fluxes
    def propertyType(self):
        return self._propertyType
    def outputs(self):
        return self._outputs

    # "Call" method creates a property instance from a registration.
    # This is the only way to create a property instance, and this
    # routine does not do any book-keeping with the AllProperties
    # object.  We do not use the Registration class's __call__ method,
    # because we need to pass the registration itself as an argument to
    # the property constructor.
    def __call__(self):
        # 2nd argument is the name.
        return self.subclass(self, self._name, *[p.value for p in self.params])


    # Collision looks for the name under this tree, and if it finds
    # it, checks if the parameter values are all equal.  If a
    # collision occurs and the parameters conflict, an exception is
    # raised.  If collision returns "true", that means a collision
    # occurred but the parameters matched.  If collision returns
    # "false", a collision did not occur.  "name" argument is the
    # trailing part of the name.  Called by
    # PropertyManager.creatorcallback, which is the callback for
    # OOF.LoadData.Property.

    # Return value is (namecollision, parametercollision)
    def collision(self, name, params):
        if name == "":
            for ppair in zip(params, self.params):
                if ppair[0].value != ppair[1].value:
                    return (1, 1)
            return (1, 0)
        else: # Nontrivial name, look for it in the local subtree.
            ltn = AllProperties.data.reverse_dict[self]
            for other in [x.object for x in ltn.nodes]:
                if other.moniker==name:
                    # "Other" params list starts with a name, because
                    # it's a NamedPropertyRegistration -- don't compare that
                    # one, passed-in params omits it.
                    for ppair in zip(params, other.params[1:]):
                        if ppair[0].value != ppair[1].value:
                            return (1, 1)
                    return (1, 0) # Collision occurred, parameters matched.
            return (0, 0) # Made it through "other", no collision.


    # Called, locally via the init or from the menucallback function.
    # (The GUI calls the menu when the dialog box closes).  Builds
    # new property instances for all the materials it knows about.
    def new_params(self):
        # "m" is a MaterialProps object, "o" is the old property instance.
        for (m, o) in self.materials.values():
            newcopy = self() # Run the registration to get the prop.
            m.new_params(o, newcopy)
            self.materials[m.name]=(m, newcopy)
            switchboard.notify("material changed", m.name)
        
    def __repr__(self):
        return "PropertyRegistration(%s, %s, %s)" % \
               (self.subclass.__name__, `self.ordering`,
                `self.params`)

    def writeData(self, datafile):
        datafile.startCmd(
            AllProperties.data.reverse_dict[self].\
            getOOFMenu(AllProperties.loadkey))
        datafile.argument('name', '')
        for param in self.params:
            datafile.argument(param.name, param.value)
        datafile.endCmd()

#### Redundant function.  Use name() instead.
##    def path(self):
##        return AllProperties.data.reverse_dict[self].path()

    def getDefaultParams(self):
        return self.params[:]

##    # Copy the values out of the incoming parameter set into
##    # the local parameter set.
##    def setDefaultParamsFromParams(self, params):
##        dct = {}
##        for incoming in params:
##            debug.fmsg(incoming.name, '=', incoming.value)
##            dct[incoming.name] = incoming.value
##        for local in self.params:
##            try:
##                local.value = dct[local.name]
##            except KeyError:
##                pass # Incomplete specification is OK.
        
    # Create a NamedPropertyRegistration from yourself.
    # Called from PropertyManager.new_prop() and
    # PropertyManager.creatorcallback()
    def named_copy(self, name, instance_params=[]):
        old_params = instance_params or self.params
        new_params = [parameter.StringParameter('name',name)] + \
                     [p.clone() for p in old_params]
        return NamedPropertyRegistration(self, name, self.subclass,
                                         self.modulename, 
                                         self.ordering, new_params)
    
    
# Registration object for named properties.  Conventionally,
# all named properties have an "ordering" equal to that of their
# parent.  They differ from the PropertyRegistration class in
# being created automatically by the tree, and by having a "parent"
# attribute which refers to a PropertyRegistration object.
class NamedPropertyRegistration(PropertyRegistration):
    def __init__(self, parent, name, subclass,
                 modulename, ordering, params):
        
        PropertyRegistrationParent.__init__(self, subclass,
                                            modulename, ordering)

        self._name = name
        self.moniker = string.split(name,":")[-1]
        self.parent = parent
        self.materials = {}
        self.params = params
        # At creation-time, all params are new.
        self.new_params()

        # Add yourself to the AllProperties object.  "name"
        # should be a fully qualified name.
        AllProperties[name] = self

    def baseName(self):
        return self.parent.baseName()

    def fields(self):
        return self.parent.fields()
    def fluxes(self):
        return self.parent.fluxes()
    def propertyType(self):
        return self.parent.propertyType()
    def outputs(self):
        return self.parent.outputs()
        
    # We do not expect NamedPropertyRegistrations to be used
    # for anything other than direct instantiation via the pre-set
    # params, so no kwargs for this one.  2nd argument is name.
    def __call__(self):
        return self.subclass(self, self._name, # self.moniker,
                             *[p.value for p in self.params[1:]])

    def __repr__(self):
        return "NamedPropertyRegistration(%s, %s, %s, %s)" % \
               (`self.moniker`, self.subclass.__name__, `self.ordering`,
                `self.params`)

    def writeData(self, datafile):
        datafile.startCmd(
            AllProperties.data.reverse_dict[self.parent].\
            getOOFMenu(AllProperties.loadkey))
        
        # Use only the tail of the fully-qualified name in the menu,
        # for brevity.  The menu item implicitly knows the rest of the
        # name.
        datafile.argument('name', self.moniker)
        for param in self.params[1:]:
            datafile.argument(param.name, param.value)
        datafile.endCmd()

    # Similar to parent "named_copy", except it gets the ordering from
    # the parent.
    def named_copy(self, name, instance_params = []):
        # Don't clone the name parameter, since the new property has a new name.
        non_name_params = instance_params or self.params[1:]
        new_params = [parameter.StringParameter('name', name)] + \
                     [p.clone() for p in non_name_params]
        return NamedPropertyRegistration(self.parent, name, self.subclass,
                                         self.modulename, 
                                         self.parent.ordering, new_params)
    
    # Return all except the "name" parameter, which is fixed by
    # the tree structure.
    def getDefaultParams(self):
        return self.params[1:]
