# -*- python -*-
# $RCSfile: rationalize.py,v $
# $Revision: 1.69 $
# $Author: haan $
# $Date: 2004/07/23 21:39:22 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


## NOTE: Each RationalizeMethod is initialized with an internal
## gerund parameter. This parameter reports via the progress bar
## (through the progress bar message) the gerund of the action
## (rationalize method) that is being executed. 

from SWIG.common import switchboard
from SWIG.common.IO import progressbar
from common import debug
from common import registeredclass
from common.IO import parameter
from common.IO import reporter
from engine import rationalizefunctions
from engine import skeletonmodifier
import random

##################################################

# Subclasses of Rationalizer must have a function "findAndFix"
# which takes a skeleton and an element as arguments and returns a
# list of ProvisionalChanges objects indicating what can be done.
# ProvisionalChanges is defined in skeleton.py. (Skeleton.mergeNode()
# returns one.)

class Rationalizer(registeredclass.RegisteredClass):
    registry = []

    def __call__(self, skel, context, targets, criterion):
        pBar = progressbar.getcProgressBar()
        # Copy the element list from skeleton.element before
        # rationalizing. This is necessary because rationalizing
        # modifies the Skeleton's element list. 
        elements = targets(skel, context, copy=1)
        random.shuffle(elements)
        executed_action = self.getRegistration().gerund
        processed = {}
        count = 0
        done = 0  # No. of rationalized elements
        nel = len(elements)  # No. of elements in the list
        for element in elements:
            count += 1  # The i-th element being processed ... for progress bar
            if not processed.has_key(element) and element.active(skel):
                # The derived class's findAndFix routine returns a
                # list of ProvisionalChanges objects, from which we
                # pick the best one.
                changes = self.findAndFix(skel, element)
                bestchange = criterion(changes, skel)
                if bestchange is not None:
                    done += bestchange.nRemoved()
                    # Accepting the change converts provisional
                    # elements to actual elements.
                    bestchange.accept(skel)
                    for elephant in bestchange.removed:
                        processed[elephant] = 1
                    for oldel, newel in bestchange.substitutions:
                        # If an unprocessed element has been replaced,
                        # its replacement still has to be processed.
                        # The element being replaced should *not* be
                        # processed, in any case, since it's no longer
                        # in the skeleton.
                        # If the criterion is "Unconditional" or
                        # "Limited Unconditional", it doesn't add subs to
                        # the list.
                        if not processed.has_key(oldel):
                            processed[oldel] = 1
                            if criterion.addSubstitute(elements, newel):
                                nel += 1

            if pBar.query_stop():
                pBar.set_failure()
                pBar.set_message("Failed")
                break
            else:
                pBar.set_progress(1.0*count/nel)
                pBar.set_message(executed_action + ": %d/%d" % (count, nel))
        skel.cleanUp()
        
        reporter.report("%d elements rationalized : %s."
                        % (done, self.getRegistration().name()))

################################################################

## In order for the widget for the list of RationalizeMethods to come
## up by default with all of the methods enabled, the "methods"
## parameter needs to be set to a list containing an instance of each
## RationalizeMethod.  This list is constructed here, and it's updated
## each time a new RationalizeMethod is created.  The update is done
## carefully, so that the id of the list doesn't change, so that the
## RegisteredListParameter still refers to the correct list.
    
def _updateDefaults():
    # rebuild list in place.  Don't just do "defaults = [...]" !
    global defaults
    del defaults[:]                   
    defaults += [reg() for reg in Rationalizer.registry]

defaults = []
_updateDefaults()
switchboard.requestCallback(Rationalizer, _updateDefaults)

################################################################

class RationalizeMethod(registeredclass.RegisteredClass):
    registry = []

removeBadTriangle = rationalizefunctions.removeBadTriangle
removeShortSide = rationalizefunctions.removeShortSide
quadSplit = rationalizefunctions.quadSplit

class Automatic(RationalizeMethod):
    def rationalize(self, skel, context, targets, criterion):
        pBar = progressbar.getcProgressBar()
        # Copy the element list from skeleton.element before
        # rationalizing.  This is necessary because rationalizing
        # modifies the Skeleton's element list. 
        elements = targets(skel, context, copy=1)
        random.shuffle(elements)
        processed = {}
        count = 0
        done = 0
        nel = len(elements)
        for element in elements:
            count += 1
            if not processed.has_key(element) and element.active(skel):
                # If the element is a triangle ...
                if element.nnodes() == 3:
                    obtuse_angle = element.getBiggestAngle()
                    acute_angles = [(obtuse_angle+1)%3, (obtuse_angle+2)%3]
                    changes = removeBadTriangle(skel, element,
                                                obtuse_angle, acute_angles)
                # If it's a quad ...
                elif element.nnodes() == 4:
                    shortEdge = element.getShortestEdge()
                    bigAngle = element.getBiggestAngle()
                    changes = removeShortSide(skel, element, shortEdge) + \
                              quadSplit(skel, element, bigAngle)
                bestchange = criterion(changes, skel)
                if bestchange is not None:
                    done += bestchange.nRemoved()
                    # Accepting the change converts provisional
                    # elements to actual elements.
                    bestchange.accept(skel)
                    for elephant in bestchange.removed:
                        processed[elephant] = 1
                    for oldel, newel in bestchange.substitutions:
                        # If an unprocessed element has been replaced,
                        # its replacement still has to be processed.
                        # The element being replaced should *not* be
                        # processed, in any case, since it's no longer
                        # in the skeleton.
                        # If the criterion is "Unconditional" or
                        # "Limited Unconditional", it doesn't add subs to
                        # the list.
                        if not processed.has_key(oldel):
                            processed[oldel] = 1
                            if criterion.addSubstitute(elements, newel):
                                nel += 1
                processed[element] = 1
                
            if pBar.query_stop():
                pBar.set_failure()
                pBar.set_message("Failed")
                break
            else:
                pBar.set_progress(1.0*count/nel)
                pBar.set_message("Rationalizing " + ": %d/%d" % (count, nel))
        skel.cleanUp()
        
        reporter.report("%d elements rationalized : %s."
                        % (done, self.getRegistration().name()))
    
registeredclass.Registration(
    'Automatic', RationalizeMethod, Automatic, ordering=1,
    tip = 'Automatically find and fix the target elements. CAUTION: This option may work better with a selection of (badly shaped) elements than with an entire skeleton.') 

class Specified(RationalizeMethod):
    def __init__(self, rationalizers):
        self.rationalizers = rationalizers
    def rationalize(self, skel, context, targets, criterion):
        pBar = progressbar.getcProgressBar()
        pBar.set_message("rationalizing...     ")
        for rationalizer in self.rationalizers:
            rationalizer(skel, context, targets, criterion)
            if pBar.query_stop():  ## or pBar.get_success() <0 :
                pBar.set_failure()
                pBar.set_message("Failed")
                break
        if pBar.query_stop():  ## or pBar.get_success() <0:
            pBar.set_failure()
            pBar.set_message("Failed")
            return None
    
registeredclass.Registration(
    'Specified', RationalizeMethod, Specified, ordering=0,
    params = [
    parameter.RegisteredListParameter('rationalizers',
                                      Rationalizer,
                                      value=defaults,
                                      tip = 'Which rationalizer(s) to use.')
    ],
    tip = 'Find and fix the target elements with specified methods.')

################################################################

class Rationalize(skeletonmodifier.SkeletonModifier):
    def __init__(self, targets, criterion, method):
        self.targets = targets      # all or selected
        self.criterion = criterion    # unconditional or conditional
        self.method = method            # automatic or specified

    def apply(self, oldskeleton, context):
        skel = oldskeleton.properCopy()
        self.method.rationalize(skel, context,
                                self.targets, self.criterion)
        skel.cleanUp()
        return skel

#########################

registeredclass.Registration(
    'Rationalize', skeletonmodifier.SkeletonModifier,
    Rationalize,
    ordering=5,
    params=[
    parameter.Parameter('targets', skeletonmodifier.SkelModTargets,
                        tip = 'Which elements to modify.'),
    parameter.Parameter('criterion', skeletonmodifier.SkelModCriterion,
                        tip = 'Acceptance criterion'),
    parameter.Parameter('method',
                        RationalizeMethod,
                        tip = 'Methods for rationalization.')
    ],
    tip = 'Tools to clean up badly shaped elements.')
